import java.io.IOException;
import java.io.InputStream;

public class SoundManager extends Thread {
	private static final int MAX_ENGINE_PITCH_LOG2 = 8;
	public static final int CREAK = 0;
	public static final int HIT_CAR = 1;
	public static final int OFF_ROAD = 2;
	public static final int SMASH = 3;
	public static final int GROUNDED = 4;
	public static final int SCRATCH = 5;
	/** Sample rate of audio resources */
	private static int SAMPLE_RATE = 11025;
	/** Buffer for engine sound. int type to speed */
	private static int[] mEngine = null;
	/** Buffer for all sounds. int type to speed */
	private static int[][] mSoundClips = null;
	/** Thread state. true: work, false: break */
	private static boolean mIsWork = false;
	/** Sound buffer for pithed engine sound data */
	private static int mEngineSoundBuff[] = null;
	/** Engine sound sample count in bytes after pitching x2 */
	private static int mEngineSoundBytesCount_x2 = 0;
	/** Current sound effect id */
	private static int mCurrentSoundIndex = SCRATCH;
	/** true if need loop sound */
	private static boolean mIsLoop = false;
	/** Buffer to send sound to the DataLine */
	private static byte[] mMixBuff = null;
	/** Gain for left channel, 256 is maximum, 0 is mute */
	private static int mGainLeft_fp8 = 0;
	/** Gain for right channel, 256 is maximum, 0 is mute */
	private static int mGainRight_fp8 = 0;
	/** Current mix position */
	private static int mMixPosition = 0;
	/** Flag if manager was created */
	private static SoundManager mInstance = null;
	/** Private to make only one manager */
	private SoundManager() {
	}
	/** Load sound resource data and convert it to 16 bits sound */
	private static int[] getClip(final String name) throws IOException {
		final int HEADER_SIZE = 44;
		InputStream is = SoundManager.class.getResourceAsStream( "/audio/" + name + ".wav" );
		try {
			if( is.available() <= HEADER_SIZE ) throw new IOException("Incorrect size for: " + name);
			int count = HEADER_SIZE + 1;
			int sample;
			while( --count > 0 ) {
				is.read();
			}
			int[] data = new int[is.available()];
			while( (sample = is.read()) >= 0 ) {// unsigned 8 to signed 16
				data[count++] = (sample - 128) << 7;
			}
			return data;
		} catch(Exception e) {
			throw new IOException( e.getMessage() );
		} finally {
			if( is != null ) is.close();
		}
	}
	/** Initiate the sound manager */
	public static int init() {
		if( mInstance != null ) {
			return -1;
		}
		mInstance = new SoundManager();
		//mInstance.setName("Sound");
		try {
			mSoundClips = new int[6][];
			//
			mEngine = getClip( "Engine" );
			mSoundClips[0] = getClip( "Creak" );
			mSoundClips[1] = getClip( "HitCar" );
			mSoundClips[2] = getClip( "OffRoad" );
			mSoundClips[3] = getClip( "Smash" );
			mSoundClips[4] = getClip( "Grounded" );
			mSoundClips[5] = getClip( "Scratch" );
			// uses pitch only >= 1
			mEngineSoundBuff = new int[mEngine.length];
			mEngineSoundBytesCount_x2 = mEngineSoundBuff.length << 2;// 2 bytes per sample * 2channels
			mMixBuff = new byte[mEngineSoundBytesCount_x2];
			return restart();
		}
		catch( Exception e ){
			System.out.print( e.getMessage() );
			release();
		}
		return -1;
	}
	// --- to pitch engine sound
	private static final int ACCURACY = 16;
	private static final int MASK = (1 << ACCURACY) - 1;
	private static int mDTf = 0;
	private static int mDT = 0;
	private static int mTf = 0;
	private static int mT = 0;
	private static int mInT = 0;
	private static int mPrevious = 0;
	/**
	 * Pitch sound.
	 * @param inSnd - input sound buffer
	 * @param outSnd - buffer with pitched sound
	 * @param inSampleRate - numerator to calculate the pitch ratio
	 * @param outSampleRate - denominator to calculate the pitch ratio
	 * @return
	 */
	private static int pitch(final int[] inSnd, final int[] outSnd, int inSampleRate, int outSampleRate) {
		final int quotient = (int)(((long)inSampleRate << ACCURACY) / (long)outSampleRate);
		mDTf = (quotient & MASK);
		mDT = (quotient >> ACCURACY);
		int out_count = 0;
		int s = 0;
		do {
			while( mInT <= mT ) {
				mPrevious = inSnd[s++];
				mInT++;
				if( s >= inSnd.length ) {
					if( mInT < mT ) {
						mT = mT - mInT;
						mInT = 0;
					} else {
						mInT = mInT - mT;
						mT = 0;
					}
					return out_count;
				}
			}
			int t = (mPrevious + ((mTf * (inSnd[s] - mPrevious)) >> ACCURACY));
			outSnd[out_count++] = t;
			t = mTf + mDTf;
			mT = mT + mDT + (t >> ACCURACY);
			mTf = (t & MASK);
		}
		while( out_count < outSnd.length );
		if( mInT < mT ) {
			mT = mT - mInT;
			mInT = 0;
		} else {
			mInT = mInT - mT;
			mT = 0;
		}
		return out_count;
	}
	// --- end pitch engine sound
	/**
	 * Use preresampled engine sounds if using system resampler with limitations
	 * @param revs
	 */
	/*public static void engine(int revs) {
		int engine_sound_index = 0;
		int frequency = (revs << (MAX_ENGINE_PITCH_LOG2 - 2)) + SAMPLE_RATE;
		//frequency |= GameView.mRandom.nextInt( 512 );// fluctuation
		frequency = Math.min( frequency, (1 << MAX_ENGINE_PITCH_LOG2) * SAMPLE_RATE );
		
		while( frequency >= 2 * SAMPLE_RATE ) {
			frequency >>= 1;
			if( engine_sound_index < (MAX_ENGINE_PITCH_LOG2 - 1) ) engine_sound_index++;
		}
		mEngineSoundBytesCount_x2 = (pitch( mSoundClips[engine_sound_index], mEngineSoundBuff, frequency, SAMPLE_RATE )) << 2;
	}*/
	/**
	 * Generate engine sound by pitching
	 * @param revs - engine revolution count
	 * @return mEngineSoundBuff - fills by the sound data
	 * @return mEngineSoundBytesCount_x2 - sound bytes count multiplied by 2
	 */
	public static void engine(int revs) {
		int frequency = (revs << (MAX_ENGINE_PITCH_LOG2 - 2)) + SAMPLE_RATE;
		//frequency |= GameView.mRandom.nextInt( 512 );// fluctuation
		frequency = Math.min( frequency, (1 << MAX_ENGINE_PITCH_LOG2) * SAMPLE_RATE );
		//mEngineSoundBytesCount_x2 = (pitch( mEngine, mEngineSoundBuff, frequency, SAMPLE_RATE )) << 2;
	}
	/** Stop sound and close audio line */
	public static void release() {
		mIsWork = false;
	}
	/**
	 * Play selected sound once.
	 * @param sound_id - id of the preloaded sound
	 * @param gainLeft_fp8 - gain to pan left
	 * @param gainRight_fp8 - gain to pan right
	 */
	public static void play(int sound_id, int gainLeft_fp8, int gainRight_fp8) {
		mCurrentSoundIndex = sound_id;
		mGainLeft_fp8 = gainLeft_fp8;
		mGainRight_fp8 = gainRight_fp8;
		mMixPosition = 0;
		mIsLoop = false;
	}
	/**
	 * Play selected sound in loop.
	 * @param sound_id - id of the preloaded sound
	 * @param gainLeft_fp8 - gain to pan left
	 * @param gainRight_fp8 - gain to pan right
	 */
	public static void playLoop(int sound_id, int gainLeft_fp8, int gainRight_fp8) {
		mCurrentSoundIndex = sound_id;
		mGainLeft_fp8 = gainLeft_fp8;
		mGainRight_fp8 = gainRight_fp8;
		mMixPosition = 0;
		mIsLoop = true;
	}
	/** Stop play a selected sound */
	public static void stop(int sound_id) {
		playLoop( sound_id, 0, 0 );
	}
	/** Stop play all sounds */
	public static void stopAll() {
		/*playLoop( mCurrentSoundIndex, 0, 0 );
		int i = mEngineSoundBuff.length;
		do{ mEngineSoundBuff[--i] = 0; } while( i > 0 );*/
	}
	/** Restart sound manager after pause */
	public static int restart() {
		if( ! mIsWork ) {
			//
			mIsWork = false;//true;
			//mInstance.start();
		}
		return 0;
	}
	/** Return next sample for selected sound to mix */
	private static int getNextSample() {
		final int[] snd = mSoundClips[mCurrentSoundIndex];
		if( ++mMixPosition >= snd.length ) {
			mMixPosition = 0;
			if( ! mIsLoop ) {// mute
				mGainRight_fp8 = mGainLeft_fp8 = 0;
			}
		}
		return (int)snd[mMixPosition];
	}
	/** Sound thread */
	public void run() {
		try{
			while( mIsWork ) {
				int i = 0;
				do {// mix 2 mono 16 bits channels to one stereo 16 bits channel
					final int sample = getNextSample();
					final int engine_sample = mEngineSoundBuff[i >> 2];
					// left channel
					int mix = engine_sample + ((mGainLeft_fp8 * sample) >> 8);
					mMixBuff[i++] = (byte)mix;
					mMixBuff[i++] = (byte)(mix >> 8);
					// right channel
					mix = engine_sample + ((mGainRight_fp8 * sample) >> 8);
					mMixBuff[i++] = (byte)mix;
					mMixBuff[i++] = (byte)(mix >> 8);
				} while( i < mEngineSoundBytesCount_x2 );
				//mLine.write( mMixBuff, 0, mEngineSoundBytesCount_x2 );
				//mLine.drain();
			}
		} catch(Exception e) {
			System.out.println( e.getMessage() );
		} finally {
			//mLine.flush();
			//mLine.close();
			//mLine = null;
		}
	}
}